/***********************************************************************
 @C-file{
    author              = "Nelson H. F. Beebe",
    version             = "3.01",
    date                = "02 May 2000",
    time                = "08:01:32 MDT",
    filename            = "tgrind.c",
    address             = "Center for Scientific Computing
			   University of Utah
			   Department of Mathematics, 322 INSCC
			   155 S 1400 E RM 233
			   Salt Lake City, UT 84112-0090
			   USA",
    telephone           = "+1 801 581 5254",
    FAX                 = "+1 801 585 1640, +1 801 581 4148",
    URL                 = "http://www.math.utah.edu/~beebe",
    checksum            = "10962 800 2209 19380",
    email               = "beebe@math.utah.edu, beebe@acm.org,
			   beebe@ieee.org (Internet)",
    codetable           = "ISO/ASCII",
    keywords            = "AmSTeX, LaTeX 2.09, LaTeX-2e, plain TeX,
			   tgrind, typeset program listing, vgrind",
    supported           = "yes",
    docstring           = "This is a top-level driver program for
			   tgrind.

			   On UNIX systems, this program can be replaced
			   by a shell script included in the
			   distribution, but on other systems, the shell
			   is too primitive to allow a complete
			   implementation of the documented behavior, so
			   a simple standalone program is called for,
			   and used by default everywhere.

			   This program is intentionally kept very
			   small, so that even in the 640KB limit of IBM
			   PC DOS, there is sufficient memory left to
			   run another copy of the shell, plus the
			   tfontedpr program, TeX, and a DVI driver.

			   This script is a direct translation of
			   tgrind.sh, whose code originally defined the
			   interface between the user-callable tgrind
			   program and the underlying tfontedpr program.

			   Because this program needs to invoke other
			   programs, it needs to know how to construct
			   command lines for them, and that job is
			   operating-system dependent.  This version has
			   support for both UNIX and IBM PC DOS, and
			   incomplete support for VAX VMS.  Look for
			   code sections inside preprocessor
			   conditionals for symbols prefixed OS_ to
			   identify the system-dependent parts.

			   In order to maximize portability of the
			   compiled program, it is assumed that other
			   required programs (awk, tex, and tfontedpr)
			   will be found in the normal search path,
			   although an installer may choose to supply
			   absolute paths to these programs at compile
			   time, because in some environments, it may be
			   preferable to do that, even though some
			   flexibility is lost.

			   For privacy, and enhanced portability,
			   tgrind's temporary files are now created in
			   the current directory, rather than in some
			   public temporary file directory.  They
			   normally have distinctive base names to
			   distinguish them from user-created files.
			   The output file names are ISO 9660-conformant
			   for maximal portability.

			   To prevent accidental overwriting of critical
			   user files, tgrind will not run if any of its
			   required output files already exist.  In
			   practice, such collisions will be rare,
			   because on most systems, tgrind will create a
			   unique file basename from the current time.
			   However, at the installer's discretion, a
			   fixed basename can also be forced at
			   compilation time, or user-provided at run
			   time.

			   These preprocessor symbols may be redefined
			   at compile time:

			       AWK
			       BASENAME
			       DEV_NULL
			       DIR
			       DVI
			       DVI_OPTIONS
			       LIBDIR
			       PRINT
			       TEX
			       TEX_OPTIONS
			       TFONTEDPR
			       TGRIND
			       TGRINDEX

			   See their defaults below for details.

			   The checksum field above contains a CRC-16
			   checksum as the first value, followed by the
			   equivalent of the standard UNIX wc (word
			   count) utility output of lines, words, and
			   characters.  This is produced by Robert
			   Solovay's checksum utility.",
 }
***********************************************************************/

#include <config.h>

#if defined(HAVE_CTYPE_H)
#include <ctype.h>
#endif

#if defined(HAVE_STDIO_H)
#include <stdio.h>
#endif

#if defined(HAVE_STDLIB_H)
#include <stdlib.h>
#endif

#if defined(HAVE_STRING_H)
#include <string.h>
#endif

#include <boolean.h>
#include <stdc.h>

#if defined(HAVE_TIME_H)
#include <time.h>
#else
extern char	*ctime ARGS((const TIME_T *));
#endif

#if !defined(HAVE_REMOVE)
#if defined(HAVE_UNLINK)
#define remove	unlink
#endif
#endif

#if defined(HAVE_STRTOUL)
#define STRTOUL strtoul
#endif

#if !defined(STRTOUL) && defined(HAVE_STRTOL)
#define STRTOUL (unsigned long)strtol
#endif

#if !defined(STRTOUL) && defined(HAVE_ATOL)
#define STRTOUL(a,b,c) ((((b) == (char**)NULL) ? 0 : (*(b) = "", 0)), (unsigned long)atol(a))
#endif

/* Strictly, we should use "&& defined(HAVE_ATOI)" here, but on Sun
SunOS 4.1.3 with CC, configure fails to set these values correctly, so
we fall back to using atoi(), knowing that the linker will catch the
error if it is truly missing. */

#if !defined(STRTOUL) /* && defined(HAVE_ATOI) */
#define STRTOUL(a,b,c) ((((b) == (char**)NULL) ? 0 : (*(b) = "", 0)), (unsigned long)atoi(a))
#endif

#if !defined(STRTOUL)
===== This_program_cannot_be_compiled___no_strtoul_or_strtol_or_atol_or_or_atoi_available =====
#endif

typedef const char* String;
static String SNULL = (String)NULL;

static const char CHAR_MINUS             = ('-');
static const char CHAR_NUL               = ('\0');

#if defined(OS_VMS) || defined(OS_PCDOS)
static const char CHAR_SLASH             = ('/');
#define isoptionchar(c) (((char)(c) == CHAR_SLASH) || ((char)(c) == CHAR_MINUS))
#endif

#if defined(OS_UNIX) || !defined(isoptionchar)
#define isoptionchar(c) ((char)(c) == CHAR_MINUS)
#endif

/* Each of the strings AWK .. TGRINDEX can be redefined at compile time */

#if !defined(AWK)
#define AWK		"awk"
#endif

#if !defined(BASENAME)
#if defined(OS_PCDOS) || defined(OS_UNIX) || defined(OS_VMS)
#define BASENAME	(SNULL)	/* SNULL ==> use `random' number in name */
#else
#define BASENAME	"tgrtmp"
#endif
#endif

#if !defined(DEV_NULL)
#if defined(OS_PCDOS)
#define DEV_NULL	"NUL"
#endif

#if defined(OS_UNIX)
#define DEV_NULL	"/dev/null"
#endif

#if defined(OS_VMS)
#define DEV_NULL	""
#endif
#endif /* !defined(DEV_NULL) */

#if !defined(DEV_NULL)
#define DEV_NULL	""
#endif

#if !defined(DIR)
#if defined(OS_PCDOS)
#define DIR		"dir"
#endif

#if defined(OS_UNIX)
#define DIR		"ls -l"
#endif

#if defined(OS_VMS)
#define DIR		"DIRECTORY /FULL"
#endif
#endif /* !defined(DIR) */

#if !defined(DIR)
#define DIR		(SNULL)
#endif

#if !defined(DVI)
#define DVI		"dvips"
#endif

#if !defined(DVI_OPTIONS)
#define DVI_OPTIONS	"-o"
#endif

#if !defined(LIBDIR)
/* NB: definitions MUST include trailing separator! */
#if defined(OS_PCDOS)
#define LIBDIR		"c:\\tgrind\\"
#endif

#if defined(OS_UNIX)
#define LIBDIR		"/usr/local/lib/tex/inputs/"
#endif

#if defined(OS_VMS)
#define LIBDIR		"SYS$TEXINPUTS:"
#endif
#endif

#if !defined(LIBDIR)
#define LIBDIR		(SNULL)
#endif

#if !defined(PRINT)
#if defined(OS_PCDOS)
#define PRINT		"print"
#endif

#if defined(OS_UNIX)
#define PRINT		"lpr"
#endif

#if defined(OS_VMS)
#define PRINT		"print"
#endif

#endif /* !defined(PRINT) */

#if !defined(PRINT)
#define PRINT		(SNULL)
#endif

#if !defined(TEX)
#define TEX		"tex"
#endif

#if !defined(TEX_OPTIONS)
#define TEX_OPTIONS	""
#endif

#if !defined(TFONTEDPR)
#define TFONTEDPR	"tfontedpr"
#endif

#if !defined(TGRIND)
#define TGRIND		"tgrind"
#endif

#if !defined(TGRINDEX)
#define TGRINDEX	"tgrindex.awk"
#endif

#define APPEND_AFTER_SPACE(list,element) append(append((list)," "),(element))

#define FPRINTF		(void)fprintf
#define SPRINTF		(void)sprintf

static String	p__;
#define MAKE_STRING(p)	(p__ = p, (p__ == SNULL) ? "" : (p__))
#define NEW_STRING(p)	append(SNULL,p)

#define TOLOWER(c) (isupper((int)(c)) ? tolower((int)(c)) : (c))

static String	append ARGS((String source, String suffix));
static void	check_args ARGS((String this_arg, String next_arg));
static void	fatal_error ARGS((String format, String value));
int		main ARGS((int argc, char* argv[]));
static void	Ssystem ARGS((String command));
static String	new_file ARGS((String basename, String extension));
static void	Sfree ARGS((String s));
static void	usage ARGS((VOID_ARG));
static void	version ARGS((VOID_ARG));

static String	Program;


#if STDC
int
main(int argc, char *argv[])
#else
int
main(argc, argv)
int argc;
char *argv[];
#endif
{
    boolean do_format_only;
    boolean do_index;
    boolean do_print;
    boolean do_right_number;
    boolean quiet;

    FILE *fpin;
    FILE *fpout;

    size_t increment;
    size_t k;

    String basename;
    String command;
    String dvi_file;
    String file_list;
    String font_family;
    String log_file;
    String ps_file;
    String tex_file;
    String tmp_file;

    TIME_T now;

    now = time((TIME_T*)NULL);

    /* Initialize to defaults documented in tgrind.man */

    if (BASENAME == SNULL)
    {
	/* For privacy reasons, we no longer put tgrind's temporary
	   files in a public temporary files directory.  Instead, we
	   create a `random' basename from the current time in seconds
	   since the epoch.  Nevertheless, we still check later and
	   terminate with a fatal error if continuing execution would
	   overwrite any existing files. */

	basename = NEW_STRING("tg_12345"); /* MS-DOS 8+3 limit */
	SPRINTF((char *)basename,"tg_%05lu",
		(unsigned long)now % (unsigned long)100000L);
    }
    else
	basename = BASENAME;

    do_format_only = FALSE;
    do_index = TRUE;
    do_print = FALSE;
    do_right_number = FALSE;
    font_family = "ComputerModern";
    increment = (size_t)10;
    quiet = FALSE;

    Program = (String)((argv[0] == (char *)NULL) ? TGRIND : argv[0]);
    file_list = SNULL;

    command = NEW_STRING(TFONTEDPR);

    for (k = 1; k < (size_t)argc; ++k)
    {
	if (isoptionchar(argv[k][0]))
	{
	    switch(TOLOWER(argv[k][1]))
	    {
	    case '?':	/* -? (help)*/
		usage();
		exit(EXIT_SUCCESS);	/* NOT REACHED */
		break;

	    case 'd':	/* -d description [tfontedpr] */
		check_args(argv[k], argv[k+1]);
		command = APPEND_AFTER_SPACE(command, argv[k]);
		command = APPEND_AFTER_SPACE(command, argv[++k]);
		break;

	    case 'f':	/* -f (force output to stdout) or -fn fontfamily */
		if (argv[k][2] == CHAR_NUL)
		    do_format_only = TRUE;
		else if ((TOLOWER(argv[k][2]) == 'n') && (argv[k][3] == CHAR_NUL))
		{
		    check_args(&argv[k][1], argv[k+1]);
		    font_family = (String)argv[++k];
		}
		else
		{
		    usage();
		    fatal_error("unrecognized argument: [%s]",argv[k]);
		}
		break;

	    case 'h':	/* -h header [tfontedpr] */
		check_args(argv[k], argv[k+1]);
		command = APPEND_AFTER_SPACE(command, argv[k]);
		command = APPEND_AFTER_SPACE(command, "\"");
		command = append(command, argv[++k]);
		command = append(command, "\"");
		break;

	    case 'i':	/* -i increment */
		check_args(argv[k], argv[k+1]);
		{
		    char *endptr;

		    increment = (size_t)STRTOUL(argv[++k], &endptr, 0);
		    if (*endptr != CHAR_NUL)
			fatal_error("option %s must have a numeric argument",argv[k-1]);
		}
		break;

	    case 'l':	/* -l language [tfontedpr] */
		check_args(argv[k], argv[k+1]);
		command = APPEND_AFTER_SPACE(command, argv[k]);
		command = APPEND_AFTER_SPACE(command, argv[++k]);
		break;

	    case 'n':	/* -n (no bold keywords) [tfontedpr] */
		check_args(argv[k], "");
		command = APPEND_AFTER_SPACE(command, argv[k]);
		break;

	    case 'o':	/* -o basename */
		check_args(argv[k], argv[k+1]);
		basename = (String)argv[++k];
		break;

	    case 'p':	/* -p (print) */
		do_print = TRUE;
		break;

	    case 'q':	/* -q (quiet) */
		quiet = TRUE;
		break;

	    case 'r':	/* -r (numbering on right) */
		do_right_number = TRUE;
		break;

	    case 'v':	/* -x (no index) */
		version();
		exit(EXIT_SUCCESS);
		break;

	    case 'x':	/* -x (no index) */
		do_index = FALSE;
		break;

	    case '-':	/* remaining arguments are all files [tfontedpr] */
		for (++k ; k < (size_t)argc; ++k)
		    file_list = APPEND_AFTER_SPACE(file_list, argv[k]);
		break;

	    default:
		usage();
		fatal_error("unrecognized option: [%s]",argv[k]);
		break;
	    }
	}
	else				/* must be file */
	    file_list = APPEND_AFTER_SPACE(file_list, argv[k]);
    }

    tex_file = new_file(basename,".tex");
    ps_file = new_file(basename,".ps");
    dvi_file = new_file(basename,".dvi");
    log_file = new_file(basename,".log");
    tmp_file = new_file(basename,".tmp");

    command = APPEND_AFTER_SPACE(command, file_list);

#if defined(OS_PCDOS) || defined(OS_UNIX)
    command = APPEND_AFTER_SPACE(command,">>");
    command = APPEND_AFTER_SPACE(command,tex_file);
#else
    fatal_error("do not know how to append output on your operating system", "");
#endif

    fpout = fopen(tex_file,"w");
    if (fpout == (FILE*)NULL)
	fatal_error("cannot open output file [%s]",tex_file);

    FPRINTF(fpout,	/* NB: no final \n: ctime() includes it! */
	    "%%%% %s version %s %s for files %s on %s",
	    MAKE_STRING(TGRIND), MAKE_STRING(VERSION),
	    MAKE_STRING(DATE), MAKE_STRING(file_list),
	    MAKE_STRING(ctime((const TIME_T*)&now)));
    Sfree(file_list);

    FPRINTF(fpout, "\\ifx \\BeginTgrind \\undefined\n");
    FPRINTF(fpout, "  \\input tgrindmac\n");
    FPRINTF(fpout, "\\fi\n");

    if ((font_family != SNULL) && (*font_family != CHAR_NUL))
	FPRINTF(fpout, "\\def \\TgrindFontFamily{%s}\n", font_family);
    else	/* show default font as a comment, but use prevailing font */
	FPRINTF(fpout, "%% \\def \\TgrindFontFamily{%s}\n",
		"ComputerModern");

    FPRINTF(fpout, "\\TgrindLineIncrement = %lu\n",
	    (unsigned long)increment);

    FPRINTF(fpout, "\\TgrindRightNumber%s\n",
	    (do_right_number ? "true" : "false"));

    (void)fclose(fpout);

    Ssystem(command);	/* tfontedpr does the rest of the work, appending to $BASENAME.tex */

    if (do_index)
    {
	command = NEW_STRING(AWK);
	command = append(command," -f ");
	command = append(command,LIBDIR);
	command = append(command,TGRINDEX);
	command = append(command," < ");
	command = append(command,tex_file);
	command = append(command," > ");
	command = append(command,tmp_file);

	Ssystem(command);

	if (remove(tex_file) != 0)
	    fatal_error("could not remove temporary file [%s]", tex_file);
	if (rename(tmp_file,tex_file) != 0)
	    fatal_error("could not rename temporary file [%s]", tmp_file);
    }

    if (do_format_only)
    {
	fpin = fopen(tex_file,"r");
	if (fpin == (FILE*)NULL)
	    fatal_error("could not reopen temporary file [%s] for input", tex_file);

	{
	    register int c;

	    while ((c = getc(fpin)) != EOF)	/* copy the temporary file to stdout */
		putchar(c);
	}

	(void)fclose(fpin);
	remove(tex_file);
    }
    else
    {
	if (!quiet)
	    (void)printf("TeXing...\n");

	command = NEW_STRING(TEX);
	if ((TEX_OPTIONS != SNULL) && (*TEX_OPTIONS != CHAR_NUL))
	    command = APPEND_AFTER_SPACE(command, TEX_OPTIONS);
	command = APPEND_AFTER_SPACE(command, tex_file);
	if (quiet)
	{
	    command = APPEND_AFTER_SPACE(command, ">");
	    command = APPEND_AFTER_SPACE(command, DEV_NULL);
	}

	Ssystem(command);

	fpin = fopen(dvi_file,"r");
	if (fpin == (FILE*)NULL)
	    fatal_error("could not open DVI file [%s] for input", dvi_file);

	if (!quiet)
	    (void)printf("\nCreating PostScript output...\n");

	if ((DVI != SNULL) && (*DVI != CHAR_NUL))
	{
	    command = NEW_STRING(DVI);
	    if (quiet)
	    {
		if (strcmp(DVI,"dvips") == 0)
		    command = APPEND_AFTER_SPACE(command, "-q");
		else if (strcmp(DVI,"dvialw") == 0)
		    command = APPEND_AFTER_SPACE(command, "-quiet=1");
	    }
	    if ((DVI_OPTIONS != SNULL) && (*DVI_OPTIONS != CHAR_NUL))
		command = APPEND_AFTER_SPACE(command, DVI_OPTIONS);
	    command = APPEND_AFTER_SPACE(command, ps_file);
	    command = APPEND_AFTER_SPACE(command, dvi_file);

	    if (quiet)
	    {
		command = APPEND_AFTER_SPACE(command, ">");
		command = APPEND_AFTER_SPACE(command, DEV_NULL);
	    }

	    Ssystem(command);
	}

	if (!quiet)
	{
	    (void)printf("\nOutput files...\n");
	    command = NEW_STRING(DIR);
	    command = APPEND_AFTER_SPACE(command, basename);
	    command = append(command, ".*");

	    Ssystem(command);
	}

	if (do_print)
	{
	    if (!quiet)
		(void)printf("\nPrinting...\n");

	    command = NEW_STRING(PRINT);
	    command = APPEND_AFTER_SPACE(command, ps_file);
	    if (quiet)
	    {
		command = APPEND_AFTER_SPACE(command, ">");
		command = APPEND_AFTER_SPACE(command, DEV_NULL);
	    }

	    Ssystem(command);
	}
    }

    Sfree(command);

    exit (EXIT_SUCCESS);
    return (EXIT_SUCCESS);
}


#if STDC
static String
append(String source, String suffix)
#else
static String
append(source, suffix)
String source;
String suffix;
#endif
{	/* return the concatenation of source and suffix, freeing source */
    size_t	need;
    char	*p;

    need = strlen(MAKE_STRING(source)) + strlen(MAKE_STRING(suffix)) + 1;

    p = (char*)malloc(need);
    if (p == (char*)NULL)
	fatal_error("out of memory","");

    p = strcat(strcpy(p,MAKE_STRING(source)), MAKE_STRING(suffix));

    Sfree(source);

    return ((String)p);
}


#if STDC
static void
check_args(String this_arg, String next_arg)
#else
static void
check_args(this_arg,next_arg)
String this_arg;
String next_arg;
#endif
{
    if (this_arg[2] != CHAR_NUL)
	fatal_error("unrecognized option: [%s]",this_arg);
    else if (next_arg == SNULL)
	fatal_error("%s option must have argument",this_arg);
}


#if STDC
static void
fatal_error(String format, String value)
#else
static void
fatal_error(format,value)
String format;
String value;
#endif
{
    FPRINTF(stderr,"%s: ",Program);
    FPRINTF(stderr,format,value);
    FPRINTF(stderr,"\n");

    exit(EXIT_FAILURE);
}


#if STDC
static String
new_file(String basename, String extension)
#else
static String
new_file(basename, extension)
String basename;
String extension;
#endif
{
    FILE *fpin;
    String filename;

    filename = append(NEW_STRING(basename),extension);

    fpin = fopen(filename,"r");
    if (fpin != (FILE*)NULL)
    {
	(void)fclose(fpin);
	fatal_error("output file [%s] already exists",filename);
    }

    return (filename);
}


#if STDC
static void
Ssystem(String command)
#else
static void
Ssystem(command)
String command;
#endif
{
			/* execute command, then free it before returning */
#if defined(DEBUG)
    FPRINTF(stderr,"DEBUG: Constructed command: [%s]\n",
	    MAKE_STRING(command));
#endif

    if (command != SNULL)
	(void)system(command);
    Sfree(command);
}


#if STDC
static void
Sfree(String s)
#else
static void
Sfree(s)
String	s;
#endif
{
    if (s != SNULL)
	free((VOIDP)s);			/* NB: cast away const */
}


static void
usage(VOID_ARG)
{
    FPRINTF(stderr,"Usage:\n");
    FPRINTF(stderr,"\t%s [ -? ] [ -d description-file ] [ -f ] [ -fn fontfamily ]\n",
	    Program);
    FPRINTF(stderr,"\t\t[ -h header ] [ -i increment ] [ -l language ] [ -n ]\n");
    FPRINTF(stderr,"\t\t[ -o basename ] [ -p ] [ -q ] [ -r ] [ -v ] [ -x ] [ - ]\n");
    FPRINTF(stderr,"\t\tfilename ...\n");
}


static void
version(VOID_ARG)
{
    FPRINTF(stderr, "%s version %s %s\n",
	    MAKE_STRING(TGRIND), MAKE_STRING(VERSION), MAKE_STRING(DATE));
}
